# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Iván Rincón 2018

from ..base.functions import hash_string
from ..base.string_checkers import field, date


class _HashGenerator:
    def __init__(self, document_number: str, birth_date: str, expiry_date: str, transliteration: dict):
        self.document_number = document_number
        self.birth_date = birth_date
        self.expiry_date = expiry_date
        self.transliteration = transliteration  # included to transliterate optional fields

    @property
    def document_number(self) -> str:
        """Return document number string

        Document number given by the issuing State or organization, to uniquely identify the document
        from all other MRTDs

        """
        return self._document_number

    @document_number.setter
    def document_number(self, value: str):
        """Set document number

        Document number given by the issuing State or organization, to uniquely identify the document
        from all other MRTDs

        Case insensitive.

        """
        self._document_number = field(value, 9, "passport number")

    @property
    def document_number_hash(self) -> str:
        """Return hash digit of document number

        """
        return hash_string(self.document_number)

    @property
    def birth_date(self) -> str:
        """Return birth date of holder

        """
        return self._birth_date

    @birth_date.setter
    def birth_date(self, value: str):
        """Set holder's date of birth with 'YYMMDD' format

        """
        self._birth_date = date(value)

    @property
    def birth_date_hash(self) -> str:
        """Return hash digit of birth date

        """
        return hash_string(self.birth_date)

    @property
    def expiry_date(self) -> str:
        """Return date of expiry of the document

        """
        return self._expiry_date

    @expiry_date.setter
    def expiry_date(self, value: str):
        """Set date of expiry of the MRTD with 'YYMMDD' format

        """
        self._expiry_date = date(value)

    @property
    def expiry_date_hash(self) -> str:
        """Return hash digit of expiry date

        """
        return hash_string(self.expiry_date)



# GNU General Public License v3.0
# -*- coding UTF-8 -*- 
# (ɔ) Iván Rincón 2018

import mrz.base.countries as countries
from mrz.base.functions import full_capitalize

dictionary = countries.english


def code_list() -> list:
    """Return a sorted list with country codes

    Returns:
        The list, alphabetically sorted, of 3-letter countries codes conform to ICAO
        specifications (also included in ISO 3166-1 alpha 3)

        """
    return [dictionary[country] for country in sorted(dictionary, key=dictionary.__getitem__)]


def countries_list():
    """Return a sorted list with country names

    Returns:
        The list of all countries and organizations recognized by ICAO, denominated
        in English, alphabetically ordered.

        """
    return [country for country in sorted(dictionary)]


def countries_code_list() -> list:
    """Return a sorted list of tuples with all country names and country codes

    Returns:
        A list, alphabetically sorted, of tuples with all countries and organizations,
        denominated in English, recognized by ICAO and their 3-letters code. The first
        element of the tuple is the country name and the second is its 3-letter code.

        """
    # return [(country, code) for country, code in sorted(dictionary.items())]
    return [country for country in sorted(dictionary.items())]


def code_country_list() -> list:
    """Return a list of tuples with all 3-letter codes and their corresponding countries names

    Returns:
        A list, alphabetically sorted, of tuples with all countries and organizations,
        denominated in English, recognized by ICAO and their 3-letters code. The first
        element of the tuple is the 3-letter code and the second is the country name.

        """
    return [(dictionary[country], country) for country in sorted(dictionary, key=dictionary.__getitem__)]


def is_country(country: str) -> bool:
    """Return True if the parameter is valid country string, False otherwise

    Returns:
        True if the parameter is a valid country name, denominated in English and
        otherwise False. Note: case insensitive method

        """
    return True if full_capitalize(country) in dictionary.keys() else False


def is_code(code: str) -> bool:
    """Return True if the parameter is a valid country code string, False otherwise

    Returns:
        True if the parameter is a valid 3-letter country code conform to ICAO
        specifications and False otherwise. Note: case insensitive method.

        """
    return True if code.upper() in dictionary.values() else False


def get_code(country: str) -> object:
    """Get the 3-letter code of a valid country name, None otherwise

    Parameter:
        Country string, denominated in English. Note: case insensitive.
    Returns:
        3-letter country code, conform to ICAO specifications, of the country string
        given as parameter and None otherwise. Note: case insensitive method.

        """
    try:
        return dictionary[full_capitalize(country)]
    except KeyError:
        return None


def get_country(code: str) -> object:
    """Get the country string of a valid 3-letter code, None otherwise

    Parameter:
        3-letter code string, conform to ICAO specifications. Note: case insensitive.
    Returns:
        Country name string, denominated in English, of the 3-letter code given as
        parameter, None otherwise.

        """
    for country, value in dictionary.items():
        if value == code.upper():
            return country
    return None


def find_country(pattern: str) -> list:
    """Search all countries that matches the given pattern

    Parameter:
        String pattern to search.
    Returns:
        A list of countries, denominated in English, matching the given pattern

        """
    return [country for country in sorted(dictionary) if pattern.lower() in country.lower()]




#!/usr/bin/python3
# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# Iván Rincón 2018

# import generator.transliterations as transliteration
#
# from base.countries_ops import *
# from .td1 import TD1CodeGenerator
# from .td2 import TD2CodeGenerator
# from .td3 import TD3CodeGenerator
#
# __all__ = ["TD1CodeGenerator", "TD2CodeGenerator", "TD3CodeGenerator", "transliteration",
#            "code_list", "countries_list", "countries_code_list", "code_country_list", "is_country",
#            "is_code", "get_code", "get_country", "find_country"]



# GNU General Public License v3.0

from ..base.functions import hash_is_ok
from ._report import _Report


class _HashChecker(_Report):
    def __init__(self, document_number: str, document_number_hash: str, birth_date: str, birth_date_hash: str,
                 expiry_date: str, expiry_date_hash: str):
        self._document_number = document_number
        self._document_number_hash = document_number_hash
        self._birth_date = birth_date
        self._birth_date_hash = birth_date_hash
        self._expiry_date = expiry_date
        self._expiry_date_hash = expiry_date_hash

    @property
    def document_number_hash(self) -> bool:
        """Return True if the hash of the document number is validated, False otherwise."""

        return self._report("document number hash",
                            hash_is_ok(self._document_number, self._document_number_hash))

    @property
    def birth_date_hash(self) -> bool:
        """Return True if the hash of the birth date is validated, False otherwise."""

        return self._report("birth date hash", hash_is_ok(self._birth_date, self._birth_date_hash))

    @property
    def expiry_date_hash(self) -> bool:
        """Return True if the hash of the expiry date is validated, False otherwise."""

        return self._report("expiry date hash", hash_is_ok(self._expiry_date, self._expiry_date_hash))

    def _all_hashes(self):
        return (self.document_number_hash &
                self.birth_date_hash &
                self.expiry_date_hash)

    def __repr__(self):
        return self._all_hashes()



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# Iván Rincón 2018

from ..base.countries_ops import *
from ..base.functions import hash_string, transliterate
from ._fields import _Fields
from ._hash_fields import _HashGenerator
from ._holder_name import _HolderName

import mrz.base.string_checkers as check
import mrz.generator._transliterations as dictionary

__all__ = ["TD3CodeGenerator", "dictionary", "code_list", "countries_list", "countries_code_list",
           "code_country_list", "is_country", "is_code", "get_code", "get_country", "find_country"]


class _TD3HashGenerator(_HashGenerator):

    @property
    def optional_data(self) -> str:
        """Return optional data field (ID Number of the passport holder)

        Optional data at the discretion of the issuing State or organization. Non mandatory field.
        29 to 42 char position of the second line.

        """
        return self._optional_data

    @optional_data.setter
    def optional_data(self, value: str):
        """Set optional data field (ID Number of the passport holder)

        Optional data at the discretion of the issuing State or organization. Non mandatory field.
        29 to 42 char position of the second line.

        Case insensitive property

        """
        self._optional_data = check.field(transliterate(value, self.transliteration), 14, "optional data", "<")

    @property
    def optional_data_hash(self) -> str:
        """Return hash digit of the id_number field

        """
        return hash_string(self.optional_data)

    @property
    def final_hash(self) -> str:
        """Return final hash digit of the passport or other TD3

        """
        final_string = (self.document_number +
                        self.document_number_hash +
                        self.birth_date +
                        self.birth_date_hash +
                        self.expiry_date +
                        self.expiry_date_hash +
                        self.optional_data +
                        self.optional_data_hash)

        return hash_string(final_string)


class _TD3HolderName(_HolderName):
    def __init__(self, surname: str, given_names: str, transliteration=dictionary.latin_based()):
        _HolderName.__init__(self, surname, given_names, transliteration)

    @property
    def identifier(self) -> str:
        """Return identifier (the primary and secondary identifiers)

        """
        return check.field(self.surname + "<<" + self.given_names, 39, "full name", "<")


class TD3CodeGenerator(_Fields, _TD3HashGenerator, _TD3HolderName):
    """Calculate the string code of the machine readable zone for official travel documents of size 3 (passports)

    Params:
        document_type    (str):  Normally 'P' for passport
        country_code     (str):  3 letters code (ISO 3166-1) or country name (in English)
        surname          (str):  Primary identifier(s)
        given_names      (str):  Secondary identifier(s)
        passport_number  (str):  Passport number
        nationality      (str):  3 letters code (ISO 3166-1) or country name
        birth_date       (str):  YYMMDD
        sex              (str):  Genre. Male: 'M', Female: 'F' or Undefined 'X'
        expiry_date      (str):  YYMMDD
        id_number        (str):  Personal number. In some countries non-mandatory field. Empty string by default
        transliteration (dict):  Transliteration dictionary for non-ascii chars. Latin based by default
        force           (bool):  Disables checks for country, nationality and document_type fields.
                                 Allows to use 3-letter-codes not included in the countries dictionary
                                 and to use document_type codes without restrictions.

    """
    def __init__(self,
                 document_type: str,
                 country_code: str,
                 surname: str,
                 given_names: str,
                 passport_number: str,
                 nationality: str,
                 birth_date: str,
                 sex: str,
                 expiry_date: str,
                 optional_data="",
                 transliteration=dictionary.latin_based(),
                 force=False):
        _TD3HolderName.__init__(self, surname, given_names, transliteration)
        self.force = force
        self.document_type = document_type
        self.country_code = country_code
        self.document_number = passport_number
        self.nationality = nationality
        self.birth_date = birth_date
        self.sex = sex
        self.expiry_date = expiry_date
        self.optional_data = optional_data

    def _line1(self):
        return (self.document_type +
                self.country_code +
                self.identifier)

    def _line2(self):
        return (self.document_number +
                self.document_number_hash +
                self.nationality +
                self.birth_date +
                self.birth_date_hash +
                self.sex +
                self.expiry_date +
                self.expiry_date_hash +
                self.optional_data +
                self.optional_data_hash +
                self.final_hash)

    def __str__(self):
        return (self._line1() + "\n" +
                self._line2())



# GNU General Public License v3.0

default = "String was not recognized as a valid "


class FieldError(ValueError):
    def __init__(self, msg="", cause=""):
        self.cause = cause
        self.msg = msg or "String contains invalid characters"
        self.args = (self.msg, self.cause)
        super(FieldError, self).__init__(self.msg, self.cause)

    def __str__(self):
        return str(self.args)


class DateError(FieldError):
    def __init__(self, msg="", cause=""):
        self.msg = msg or "%sdate. It should be 'YYMMDD'" % default
        self.cause = cause
        super(DateError, self).__init__(self.msg, self.cause)


class SexError(FieldError):
    def __init__(self, msg="", cause=""):
        self.msg = msg or "%sgenre. Sex code should be 'M', 'F' or 'X'" % default
        self.cause = cause
        super(SexError, self).__init__(self.msg, self.cause)


class CountryError(FieldError):
    def __init__(self, msg="", cause=""):
        self.msg = msg or ("%scountry code or country or country name. It should be a valid co"
                           "untry code (3 letters) or a valid country name (english)" % default)
        self.cause = cause
        super(CountryError, self).__init__(self.msg, self.cause)


class DocumentTypeError(FieldError):
    def __init__(self, msg="", cause=""):
        self.msg = msg or "%stype of document." % default
        self.cause = cause
        super(DocumentTypeError, self).__init__(self.msg, self.cause)


class LengthError(FieldError):
    def __init__(self, msg="", cause=0, document="", length=0, amx=False):
        self.document = document
        self.cause = str(cause)
        extra = " as maximum" if amx else ""
        self.msg = msg or "%s%s. It should have %d characters%s" % (default, self.document, length, extra)
        super(LengthError, self).__init__(self.msg, self.cause)



# GNU General Public License v3.0


class _Report:

    _rep = [[], [], []]  # [[fields], [warnings], [errors]]

    def _report(self, description, result=None, kind=0):
        # kind = 0: fields, 1: warning, 2: error
        if result is not None:
            self._rep[0].append((description, result))
            if result is False:
                self._rep[2].append(str(result).lower() + " " + description)
        else:
            self._rep[kind].append(description)
        return result

    def _report_reset(self):
        self._rep[0] = []
        self._rep[1] = []
        self._rep[2] = []

    @property
    def report(self) -> list:
        """Returns a list with all fields checked.

        Returns:
            A list of tuples. In the tuples, the first item is a string with the field description,
            the second one is a bool with the result.

            """
        return self._rep[0]

    @property
    def report_falses(self) -> list:
        """Returns a list of tuples with all wrong fields checked."""

        return [item for item in self._rep[0] if not item[1]]

    @property
    def report_warnings(self) -> list:
        """Returns a list of detected warnings"""

        return self._rep[1]

    @property
    def report_errors(self) -> list:
        """Returns a list of detected errors"""

        return self._rep[2]



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# Iván Rincón 2018

from ..base.countries_ops import *
from ..base.functions import hash_string, transliterate
from ._fields import _Fields
from ._hash_fields import _HashGenerator
from ._holder_name import _HolderName

import mrz.generator._transliterations as dictionary
import mrz.base.string_checkers as check

__all__ = ["TD1CodeGenerator", "dictionary", "code_list", "countries_list", "countries_code_list",
           "code_country_list", "is_country", "is_code", "get_code", "get_country", "find_country"]


class _TD1HashGenerator(_HashGenerator):
    @property
    def optional_data1(self) -> str:
        """Return optional data field of the first line (16 to 30 char position of the first line) for TD1

        Optional data at the discretion of the issuing State or organization. Non mandatory field

        """
        return self._optional_data1

    @optional_data1.setter
    def optional_data1(self, value: str):
        """Set optional data field of the first line (16 to 30 char position of the first line) for TD1

        Optional data at the discretion of the issuing State or organization. Non mandatory field

        Case insensitive

        """
        self._optional_data1 = check.field(transliterate(value, self.transliteration), 15, "optional data 1", "<")

    @property
    def optional_data2(self) -> str:
        """Return optional data field of the second line (19 to 29 char position of the second line) for TD1

        Optional data at the discretion of the issuing State or organization. Non mandatory field

        """
        return self._optional_data2

    @optional_data2.setter
    def optional_data2(self, value: str):
        """Set optional data field of the second line (19 to 29 char position of the second line) for TD1

        Optional data at the discretion of the issuing State or organization. Non mandatory field

        Case insensitive.

        """
        self._optional_data2 = check.field(transliterate(value, self.transliteration), 11, "optional data 2", "<")

    @property
    def final_hash(self) -> str:
        """Return final hash digit for TD1

        """
        return hash_string(self.document_number +
                           self.document_number_hash +
                           self.optional_data1 +
                           self.birth_date +
                           self.birth_date_hash +
                           self.expiry_date +
                           self.expiry_date_hash +
                           self.optional_data2)


class _TD1HolderName(_HolderName):
    def __init__(self, surname: str, given_names: str, transliteration: dict):
        _HolderName.__init__(self, surname, given_names, transliteration)

    @property
    def identifier(self) -> str:
        """Return identifier (sum of the primary and secondary identifier)

        """
        return check.field(self.surname + "<<" + self.given_names, 30, "full name", "<")


class TD1CodeGenerator(_TD1HolderName, _TD1HashGenerator, _Fields):
    """Calculate the string code of the machine readable zone for official travel documents of size 1

    Params:
        document_type    (str):  The first letter shall be 'I', 'A' or 'C'
        country_code     (str):  3 letters code (ISO 3166-1) or country name (in English)
        document_number  (str):  Document number
        birth_date       (str):  YYMMDD
        sex              (str):  Genre. Male: 'M', Female: 'F' or Undefined 'X'
        expiry_date      (str):  YYMMDD
        nationality      (str):  3 letters code (ISO 3166-1) or country name (in English)
        surname          (str):  Holder primary identifier(s)
        given_names      (str):  Holder secondary identifier(s)
        optional_data1   (str):  Optional personal data at the discretion of the issuing State.
                                 Non-mandatory field. Empty string by default
        optional_data2   (str):  Optional personal data at the discretion of the issuing State.
                                 Non-mandatory field. Empty string by default
        transliteration (dict):  Transliteration dictionary for non-ascii chars. Latin based by default
        force           (bool):  Disables checks for country, nationality and document_type fields.
                                 Allows to use 3-letter-codes not included in the countries dictionary
                                 and to use document_type codes without restrictions.

    """
    def __init__(self,
                 document_type: str,
                 country_code: str,
                 document_number: str,
                 birth_date: str,
                 sex: str,
                 expiry_date: str,
                 nationality: str,
                 surname: str,
                 given_names: str,
                 optional_data1="",
                 optional_data2="",
                 transliteration=dictionary.latin_based(),
                 force=False):
        _TD1HolderName.__init__(self, surname, given_names, transliteration)
        self.force = force
        self.document_type = document_type
        self.country_code = country_code
        self.document_number = document_number
        self.optional_data1 = optional_data1
        self.birth_date = birth_date
        self.sex = sex
        self.expiry_date = expiry_date
        self.nationality = nationality
        self.optional_data2 = optional_data2
        self.transliteration = transliteration

    def _line1(self):
        return(self.document_type +
               self.country_code +
               self.document_number +
               self.document_number_hash +
               self.optional_data1)

    def _line2(self):
        return(self.birth_date +
               self.birth_date_hash +
               self.sex +
               self.expiry_date +
               self.expiry_date_hash +
               self.nationality +
               self.optional_data2 +
               self.final_hash)

    def __str__(self):
        return(self._line1() + "\n" +
               self._line2() + "\n" +
               self.identifier)



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

from ..base.countries_ops import *
from .td2 import TD2CodeChecker

import mrz.base.string_checkers as check


__all__ = ["MRVBCodeChecker", "code_list", "countries_list", "countries_code_list", "code_country_list",
           "is_country", "is_code", "get_code", "get_country", "find_country"]


class MRVBCodeChecker(TD2CodeChecker):
    """
    Check the string code of the machine readable visas type B (MRVB)

    __bool__() returns True if all fields are validated, False otherwise

    Params:
        mrz_string        (str):  MRZ string of td3. Must be 88 characters long (uppercase)
        check_expiry     (bool):  If it's set to True, it is verified and reported as warning that the
                                  document is not expired and that expiry_date is not greater than 10 years
        compute_warnings (bool):  If it's set True, warnings compute as False

    """
    def __init__(self, mrz_code: str, check_expiry=False, compute_warnings=False):
        TD2CodeChecker.__init__(self, mrz_code, check_expiry, compute_warnings)

    @property
    def optional_data(self) -> bool:
        """Return True if the format of the optional data field is validated, False otherwise."""

        s = self._optional_data = self.mrz_code.splitlines()[1][28: 36]
        return True if check.is_empty(s) else self._report("optional data format", check.is_printable(s))

    def _all_hashes(self) -> bool:
        return (self.document_number_hash &
                self.birth_date_hash &
                self.expiry_date_hash)



# GNU General Public License v3.0

from datetime import datetime, date, timedelta
from ._report import _Report
from ._honorifics import titles

import mrz.base.string_checkers as check


class _FieldChecker(_Report):
    def __init__(self, document_type: str, country: str, identifier: str, document_number: str, nationality: str,
                 birth_date: str, sex: str, expiry_date: str, optional_data: str, optional_data_2: str,
                 check_expiry: bool, compute_warnings: bool, mrz_code: str):
        self._compute_warnings = compute_warnings
        self._document_type = document_type
        self._country = country
        self._identifier = identifier
        self._document_number = document_number
        self._nationality = nationality
        self._birth_date = birth_date
        self._birth_date_check = True
        self._sex = sex
        self._expiry_date = expiry_date
        self._expiry_date_check = True
        self._optional_data = optional_data
        self._optional_data_2 = optional_data_2
        self._check_expiry = check_expiry
        self._mrz_code = mrz_code
        self._times()

    @property
    def mrz_code(self):
        """Return Machine Readable Zone code string"""

        return self._mrz_code

    @property
    def document_type(self) -> bool:
        """Return True if document type code is validated, False otherwise"""

        ok = False
        try:
            ok = bool(check.document_type(self._document_type, self))
        except ValueError:  # as error:
            # print("%s: %s", (error.args[0], error.args[1]))
            pass
        finally:
            return self._report("document type format", ok)

    @property
    def country(self) -> bool:
        """Return True if the country code is validated, False otherwise."""

        import mrz.base.countries_ops as chk
        return self._report("valid country code", chk.is_code(self._country.replace("<", "")))

    @property
    def identifier(self) -> bool:
        """Return True is the identifier is validated overcoming the checks, False otherwise."""

        id_ = self._identifier
        if not check.is_printable(id_):
            return self._report("identifier", False)
        ok = True
        if check.is_empty(id_):
            return self._report("empty identifier", kind=2)
        if check.uses_nums(id_):
            self._report("identifier with numbers", kind=2)
            ok &= False
        id2iter = id_.split("<<")
        if len(id2iter) == 1:
            secondary = id2iter[0]
            primary = ""
            ok &= False if self._compute_warnings else ok
            # print("Only one identifier")
            self._report("only one identifier", kind=1)
        else:
            secondary = id2iter[1]
            primary = id2iter[0]
        # print("Debug. (name, surname):", (secondary, primary))
        # secondary is not ok. secondary -= padding  # secondary.replace(padding, "")
        padding = id_[id_.find(secondary) + len(secondary):]
        padding = padding if primary != secondary else padding.replace(secondary, "")
        # print("Debug. padding:", padding)
        if not len(padding) and id_[-1] in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
            self._report("possible truncating", kind=1)
            ok &= False if self._compute_warnings else ok  # TODO: compute as False?
        if not check.is_empty(padding):
            # print("Error, Padding is not empty")
            self._report("more than two identifiers", kind=2)
            ok &= False
        if check.begin_by(primary, "<") or check.begin_by(secondary, "<"):
            self._report("some identifier begin by '<'", kind=2)
            ok &= False
        for i in range(2):
            for itm in id2iter[i].split("<"):
                if not itm:
                    pass
                else:
                    for tit in titles:
                        if tit == itm:
                            if i == 1:
                                self._report("Possible unauthorized prefix or suffix in identifier", kind=1)
                            else:
                                self._report("Possible not recommended prefix or suffix in identifier", kind=1)
                            ok &= False if self._compute_warnings else ok
        return self._report("identifier", ok)

    @property
    def document_number(self) -> bool:
        """Return True if the document number format is validated, False otherwise."""

        s = self._document_number
        return self._report("document number format",
                            not check.is_empty(s) and check.is_printable(s) and not check.begin_by(s, "<"))

    @property
    def nationality(self) -> bool:
        """
        Return True if nationality code is validated, False otherwise.

        """
        import mrz.base.countries_ops as chk
        return self._report("valid nationality code", chk.is_code(self._nationality.replace("<", "")))

    @property
    def birth_date(self) -> bool:
        """Return True is the birth date is validated, False otherwise."""

        ok = False
        try:
            # TODO: Make comment about self._birth_date_check (if check_periods == True)
            ok = False if not self._birth_date_check else bool(check.date(self._birth_date))
        except ValueError:
            pass
        finally:
            return self._report("birth date", ok)

    @property
    def sex(self) -> bool:
        """Return True if the sex is "M", "F" or "X", False otherwise."""

        ok = False
        try:
            ok = bool(check.sex(self._sex))
        except ValueError:
            pass
        finally:
            return self._report("valid genre format", ok)

    @property
    def expiry_date(self) -> bool:
        """Return True if the expiry date is validated, False otherwise."""

        ok = False
        try:
            # TODO: Make comment about self._expiry_date_check (if check_periods == True)
            ok = False if not self._expiry_date_check else bool(check.date(self._expiry_date))
        except ValueError:
            pass
        finally:
            return self._report("expiry date", ok)

    @property
    def optional_data(self) -> bool:
        """Return True if the format of the optional data field is validated, False otherwise."""

        s = self._optional_data
        return True if check.is_empty(s) else self._report("optional data format", check.is_printable(s))

    @property
    def optional_data_2(self) -> bool:
        """Return True if the format of the optional data field is validated, False otherwise."""
        s = self._optional_data_2
        return True if check.is_empty(s) else self._report("optional data 2 format", check.is_printable(s))

    def _times(self) -> bool:
        birth, expiry = "", ""

        try:
            birth = datetime.strptime(self._birth_date, "%y%m%d")
        except ValueError:
            self._birth_date_check = False

        try:
            expiry = datetime.strptime(self._expiry_date, "%y%m%d") + timedelta(days=1)
        except ValueError:
            self._expiry_date_check = False

        if self._birth_date_check & self._expiry_date_check:
            today = datetime.combine(date.today(), datetime.min.time())
            leap = not today.year % 4 and date.today() == date(today.year, 2, 29)
            birth = birth if birth < today else birth.replace(year=birth.year - 100)   # cancel check2

            check1 = expiry > birth
            # check2 = birth < today  # Canceled
            check3 = expiry > today
            today = datetime.today().replace(month=3, day=1) if leap else today
            check4 = expiry < today.replace(year=today.year + 10)

            # print("Debug:", ("Birth:", str(birth.date())), ("Expiry:", str(expiry.date())))

            rep = lambda s, c, k=2: not c and self._report(s, kind=k)
            rep("expiry date before than birth date", check1)
            # rep("birth date after than today", check2)  # check2 canceled
            self._check_expiry and rep("document expired", check3, 1)
            self._check_expiry and rep("expiry date greater than 10 years", check4, 1)

            self._birth_date_check = check1  # & check2   # check2 canceled
            self._expiry_date_check = check1 if not self._compute_warnings else check1 & check3 & check4

        return self._birth_date_check & self._expiry_date_check

    def _all_fields(self) -> bool:
        return (self.document_type &
                self.country &
                self.nationality &
                self.birth_date &
                self.expiry_date &
                self.sex &
                self.identifier &
                self.document_number &
                self.optional_data &
                self.optional_data_2)

    def __repr__(self) -> str:
        return str(self._all_fields())



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

from ..base.countries_ops import *
from ..base.functions import hash_is_ok
from ._hash_fields import _HashChecker
from ._fields import _FieldChecker

import mrz.base.string_checkers as check


__all__ = ["TD1CodeChecker", "code_list", "countries_list", "countries_code_list", "code_country_list",
           "is_country", "is_code", "get_code", "get_country", "find_country"]


class _TD1HashChecker(_HashChecker):
    def __init__(self, document_number: str, document_number_hash: str, birth_date: str, birth_date_hash: str,
                 expiry_date: str, expiry_date_hash: str, optional_data: str, optional_data_2: str,
                 final_hash: str):
        self._optional_data = optional_data
        self._optional_data_2 = optional_data_2
        self._final_hash = final_hash
        _HashChecker.__init__(self, document_number, document_number_hash, birth_date, birth_date_hash,
                              expiry_date, expiry_date_hash)

    @property
    def final_hash(self) -> bool:
        """Return True if final hash is True, False otherwise"""

        ok = hash_is_ok(self._document_number +
                        self._document_number_hash +
                        self._optional_data +
                        self._birth_date +
                        self._birth_date_hash +
                        self._expiry_date +
                        self._expiry_date_hash +
                        self._optional_data_2, self._final_hash)
        return self._report("final hash", ok)

    def _all_hashes(self) -> bool:
        return (self.final_hash &
                self.document_number_hash &
                self.birth_date_hash &
                self.expiry_date_hash)

    def __repr__(self) -> str:
        return str(self._all_hashes())


class TD1CodeChecker(_TD1HashChecker, _FieldChecker):
    """
    Check the string code of the machine readable zone for official travel documents of size 1

    __bool__() returns True if all fields are validated, False otherwise

    Params:
        mrz_string        (str):  MRZ string of td1s. Must be 90 uppercase characters long
        check_expiry     (bool):  If it's set to True, it is verified and reported as warning that the
                                  document is not expired and that expiry_date is not greater than 10 years
        compute_warnings (bool):  If it's set True, warnings compute as False

    """
    def __init__(self, mrz_code: str, check_expiry=False, compute_warnings=False):
        check.precheck("TD1", mrz_code, 90)
        lines = mrz_code.splitlines()
        self._document_type = lines[0][0: 2]
        self._country = lines[0][2: 5]
        self._identifier = lines[2]
        self._document_number = lines[0][5: 14]
        self._document_number_hash = lines[0][14]
        self._nationality = lines[1][15: 18]
        self._birth_date = lines[1][0: 6]
        self._birth_date_hash = lines[1][6]
        self._sex = lines[1][7]
        self._expiry_date = lines[1][8: 14]
        self._expiry_date_hash = lines[1][14]
        self._optional_data = lines[0][15: 30]
        self._optional_data_2 = lines[1][18: 29]
        self._final_hash = lines[1][29]
        self._report_reset()
        _TD1HashChecker.__init__(self,
                                 self._document_number,
                                 self._document_number_hash,
                                 self._birth_date,
                                 self._birth_date_hash,
                                 self._expiry_date,
                                 self._expiry_date_hash,
                                 self._optional_data,
                                 self._optional_data_2,
                                 self._final_hash)
        _FieldChecker.__init__(self,
                               self._document_type,
                               self._country,
                               self._identifier,
                               self._document_number,
                               self._nationality,
                               self._birth_date,
                               self._sex,
                               self._expiry_date,
                               self._optional_data,
                               self._optional_data_2,
                               check_expiry,
                               compute_warnings,
                               mrz_code)
        self.result = self._all_hashes() & self._all_fields()

    def __repr__(self):
        return str(self.result)

    def __bool__(self):
        return self.result



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

__author__  = "Iván Rincón"
__license__ = "GPL3"
__version__ = "0.3.5"



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

from mrz.base.functions import transliterate
from mrz.generator.td2 import TD2CodeGenerator

import mrz.base.string_checkers as check
import mrz.generator._transliterations as dictionary


class MRVBCodeGenerator(TD2CodeGenerator):
    """Calculate the string code for machine readable zone visas of type B (MRVB)

    Params:
        document_type    (str):  The First letter must be 'V'
        country_code     (str):  3 letters code (ISO 3166-1) or country name (in English)
        surname          (str):  Primary identifier(s)
        given_names      (str):  Secondary identifier(s)
        passport_number  (str):  Passport number
        nationality      (str):  3 letters code (ISO 3166-1) or country name
        birth_date       (str):  YYMMDD
        sex              (str):  Genre. Male: 'M', Female: 'F' or Undefined 'X'
        expiry_date      (str):  YYMMDD
        optional_data    (str):  Optional personal data at the discretion of the issuing State.
                                 Non-mandatory field. Empty string by default.
        transliteration (dict):  Transliteration dictionary for non-ascii chars. Latin based by default
        force           (bool):  Disables checks for country, nationality and document_type fields.
                                 Allows to use 3-letter-codes not included in the countries dictionary
                                 and to use document_type codes without restrictions.

    """

    def __init__(self,
                 document_type: str,
                 country_code: str,
                 surname: str,
                 given_names: str,
                 document_number: str,
                 nationality: str,
                 birth_date: str,
                 sex: str,
                 expiry_date: str,
                 optional_data="",
                 transliteration=dictionary.latin_based(),
                 force=False):
        TD2CodeGenerator.__init__(self, document_type, country_code, surname, given_names, document_number,
                                  nationality, birth_date, sex, expiry_date, optional_data, transliteration,
                                  force)

    @property
    def optional_data(self) -> str:
        """Return optional data field for Visas of type B (MRVB)

        Optional data at the discretion of the issuing State or organization. Non mandatory field.
        29 to 36 char position of the second line.

        """
        return self._optional_data

    @optional_data.setter
    def optional_data(self, value: str):
        """Set optional data field for Visas of type B (MRVB)

        Optional data at the discretion of the issuing State or organization. Non mandatory field.
        29 to 36 char position of the second line.

        Case insensitive property

        """
        self._optional_data = check.field(transliterate(value, self.transliteration), 8, "id number", "<")

    @property
    def final_hash(self) -> None:
        # override property (TD2)
        return None

    def _line2(self):
        return (self.document_number +
                self.document_number_hash +
                self.nationality +
                self.birth_date +
                self.birth_date_hash +
                self.sex +
                self.expiry_date +
                self.expiry_date_hash +
                self.optional_data)



# -*- coding: UTF-8 -*-

# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018


# Arabic

transliteration = {
    # ICAO specifications
    u"\u0027": "",     # '
    u"\u0628": "B",    # ب
    u"\u062B": "XTH",  # ث
    u"\u062C": "J",    # ج
    u"\u062D": "XH",   # ح
    u"\u062E": "XKH",  # خ
    u"\u062F": "D",    # د
    u"\u0630": "XDH",  # ذ
    u"\u0631": "R",    # ر
    u"\u0632": "Z",    # ز
    u"\u0633": "S",    # س
    u"\u0634": "XSH",  # ش
    u"\u0635": "XSS",  # ص
    u"\u0636": "XDZ",  # ض
    u"\u0637": "XTT",  # ط
    u"\u0638": "XZZ",  # ظ
    u"\u0639": "E",    # ع
    u"\u063A": "G",    # غ
    u"\u0641": "F",    # ف
    u"\u0642": "Q",    # ق
    u"\u0643": "K",    # ك
    u"\u0644": "L",    # ل
    u"\u0645": "M",    # م
    u"\u0646": "N",    # ن
    u"\u0647": "H",    # ه
    u"\u0648": "W",    # و
    u"\u0649": "XAY",  # ى
    u"\u064A": "Y",    # ي
    u"\u0671": "XXA",  # ٱ
    u"\u0679": "XXT",  # ٹ
    u"\u067E": "P",    # پ
    u"\u067C": "XRT",  # ټ
    u"\u0681": "XKE",  # ځ
    u"\u0685": "XXH",  # څ
    u"\u0686": "XC",   # چ
    u"\u0688": "XXD",  # ڈ
    u"\u0689": "XDR",  # ډ
    u"\u0691": "XXR",  # ڑ
    u"\u0693": "XRR",  # ړ
    u"\u0696": "XRX",  # ږ
    u"\u0698": "XJ",   # ژ
    u"\u069A": "XXS",  # ښ
    u"\u06A9": "XKK",  # ک
    u"\u06AB": "XXK",  # ګ
    u"\u06AD": "XNG",  # ڭ
    u"\u06AF": "XGG",  # گ
    u"\u06BA": "XNN",  # ں
    u"\u06BC": "XXN",  # ڼ
    u"\u06BE": "XDO",  # ھ
    u"\u06C0": "XYH",  # ۀ
    u"\u06C1": "XXG",  # ہ
    u"\u06C2": "XGE",  # ۂ
    u"\u06C3": "XTG",  # ۃ
    u"\u06CC": "XYA",  # ی
    u"\u06CD": "XXY",  # ۍ
    u"\u06D0": "Y",    # ې
    u"\u06D2": "XYB",  # ے
    u"\u06D3": "XBE",  # ۓ

    # Some experimental additions (It seems to work well)
    u"\u0621": "",     # ء
    u"\u0670": "",     # '
    u"\u0627": "A",    # ا
    u"\u064E": "A",    # َ
    u"\u064B": "F",    # ً
    u"\u0650": "I",    # ِ
    u"\u0629": "P",    # ة
    u"\u062A": "T",    # ت
    u"\u064C": "N",    # ٌ
    u"\u064D": "K",    # ٍ
    u"\u064F": "U",    # ُ
    u"\u0652": "O",    # ْ

}

# https://www.icao.int/publications/Documents/9303_p3_cons_en.pdf
# https://en.wikipedia.org/wiki/Romanization_of_Arabic



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

# from .td1 import TD1CodeChecker
# from .td2 import TD2CodeChecker
# from .td3 import TD3CodeChecker



# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Iván Rincón 2018

import mrz.base.string_checkers as check


class _Fields:
    def __init__(self, document_type: str, country_code: str, document_number: str,
                 nationality: str, sex: str, force: bool):
        self.force = force
        self.document_type = document_type
        self.country_code = country_code
        self.document_number = document_number
        self.nationality = nationality
        self.sex = sex

    @property
    def document_type(self) -> str:
        """Return document type code

        """
        return self._document_type

    @document_type.setter
    def document_type(self, value: str):
        """Set document type code

        For TD1 and TD2, the first letter must be 'I', 'A' or 'C'. Optionally, can be used a
        second character. The second character shall be at discretion of issuing state.
        Note: 'V' shall not be used as 2nd char and 'C' shall not be used after 'A'. (TD1 and TD2)

        First letter of passports and other TD3 must be 'P'. Optionally, can be used a second
        character. The second character shall be at discretion of issuing state

        For visas (MRV-A and MRV-B), the first letter must be 'V'. Optionally, can be used a
        second character. The second character shall be at discretion of issuing state

        Case insensitive.

        """
        self._document_type = check.document_type(value, self) if not self.force \
            else check.field(value, 2, "document type")

    @property
    def country_code(self) -> str:
        """Return 3-letter-code of issuing State

        Representation of issuing State or organization and nationality of holder specified in ISO 3166-1

        """
        return self._country_code

    @country_code.setter
    def country_code(self, value: str):
        """Set issuing state

        Can be used the 3-letter code (ISO 3166-1) or the country name called in English (For example,
        'NLD' or 'Netherlands')

        Case insensitive.

        """
        self._country_code = check.country(value) if not self.force else check.field(value, 3, "country code")

    @property
    def nationality(self) -> str:
        """Return 3-letter-code of nationality

        Representation of 3-letter code (ISO 3166-1) of holder's nationality

        """
        return self._nationality

    @nationality.setter
    def nationality(self, value: str):
        """Set holder's nationality

        Can be used the 3-letter code (ISO 3166-1) or the country name called in English (For example,
        'NLD' or 'Netherlands')

        Case insensitive

        """
        self._nationality = check.country(value) if not self.force else check.field(value, 3, "document type")

    @property
    def sex(self) -> str:
        """Return holder's genre

        """
        return self._sex

    @sex.setter
    def sex(self, value: str):
        """Set holder's genre

        Can be used 'F' (female), 'M' (male) or 'X' (undefined). Case insensitive.

        """
        self._sex = check.sex(value)




# -*- coding: UTF-8 -*-

# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018


# Modern greek (Experimental)

transliteration = {
    u"\u0027": "",
    u"\u0386": "A", u"\u03AC": "a",    # Ά, ά
    u"\u0391": "A", u"\u03B1": "a",    # Α, α
    u"\u0392": "B", u"\u03B2": "b",    # Β, β
    u"\u0393": "G", u"\u03B3": "g",    # Γ, γ
    u"\u0394": "D", u"\u03B4": "d",    # Δ, δ
    u"\u0395": "E", u"\u03B5": "e",    # Ε, ε
    u"\u0396": "Z", u"\u03B6": "z",    # Ζ, ζ
    u"\u0397": "I", u"\u03B7": "i",    # Η, η
    u"\u0398": "TH", u"\u03B8": "th",  # Θ, θ
    u"\u0399": "I", u"\u03B9": "i",    # Ι, ι
    u"\u039A": "K", u"\u03BA": "k",    # Κ, κ
    u"\u039B": "L", u"\u03BB": "l",    # Λ, λ
    u"\u039C": "M", u"\u03BC": "m",    # Μ, μ
    u"\u039D": "N", u"\u03BD": "n",    # Ν, ν
    u"\u039E": "X", u"\u03BE": "x",    # Ξ, ξ
    u"\u039F": "O", u"\u03BF": "o",    # Ο, ο
    u"\u03A0": "P", u"\u03C0": "p",    # Π, π
    u"\u03A1": "P", u"\u03C1": "p",    # Ρ, ρ
    u"\u03A3": "S", u"\u03C3": "s",    # Σ, σ
    u"\u03C2": "s",                    # ς
    u"\u03A4": "T", u"\u03C4": "t",    # Τ, τ
    u"\u03A5": "Y", u"\u03C5": "y",    # Υ, υ
    u"\u03A7": "CH", u"\u03C7": "ch",  # Χ, χ
    u"\u03A8": "PS", u"\u03C8": "ps",  # Ψ, ψ
    u"\u03A9": "O", u"\u03C9": "o",    # Ω, ω
    u"\u0388": "E", u"\u03AD": "e",    # Έ, έ
    u"\u0389": "I", u"\u03AE": "i",    # Ή, ή
    u"\u038A": "I", u"\u03AF": "i",    # Ί, ί
    u"\u038C": "O", u"\u03CC": "o",    # Ό, ό
    u"\u038E": "Y", u"\u03CD": "y",    # Ύ, ύ
    u"\u038F": "O", u"\u03CE": "o",    # Ώ, ώ
    u"\u03AA": "I", u"\u03CA": "i",    # Ϊ, ϊ
    u"\u03AB": "Y", u"\u03CB": "y",    # Ϋ, ϋ
}

# https://en.wikipedia.org/wiki/Romanization_of_Greek



# -*- coding: UTF8 -*-

from setuptools import setup
from os import path
from mrz import __version__ as version

parent = path.abspath(path.dirname(__file__))

with open(path.join(parent, "README.rst"), 'r') as readme:
    long_description = readme.read()

setup(
    name="mrz",
    version=version,
    description="Machine readable zone generator and checker for passports, visas, id cards and other travel documents",
    license="GPLv3",
    long_description=long_description,
    author='Ivan Rincon',
    author_email='ivan.rincon76@gmail.com',
    url="https://github.com/Arg0s1080/mrz",
    keywords="mrz passports visas id cards td1 td2 td3 mrva mrvb icao",
    packages=["mrz", "mrz.base", "mrz.checker", "mrz.generator", "mrz.generator.dictionaries"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Topic :: Security"
    ]
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# Iván Rincón 2018


def hash_string(string: str) -> str:
    """
    >>> hash_string("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
    '7'
    >>> hash_string("0123456789")
    '7'
    >>> hash_string("0123456789ABCDEF")
    '0'
    >>> hash_string("0")
    '0'
    """
    from string import ascii_uppercase
    string = string.upper().replace("<", "0")
    weight = [7, 3, 1]
    hsh = 0
    for i in range(len(string)):
        c = list(string)[i]
        w = weight[i % 3]
        if c.isdigit():
            hsh += int(c) * w
        elif c in ascii_uppercase:
            hsh += list(ascii_uppercase).index(c) * w
        else:
            raise ValueError("%s contains invalid characters" % string, c)

    return str(hsh % 10)


def hash_is_ok(string: str, hash_: str) -> bool:
    """
    >>> hash_is_ok("0123456789abcdef", "0")
    True
    >>> hash_is_ok("0", "1")
    False
    """
    return True if hash_string(string) == hash_ else False


def full_capitalize(string: str) -> str:
    """
        >>> full_capitalize("It'S A teSt")
        "It's A Test"
    """
    if " " in string:
        t = string.split()
        for i in range(len(t)):
            t[i] = t[i].capitalize()
        return " ".join(t)
    else:
        return string.capitalize()


def transliterate(string: str, dictionary: dict, sep="<") -> str:
    """
    >>> import mrz.generator.transliterations as dictionary_
    >>> transliterate("ТЕСТ МИЛИЦА", dictionary_.cyrillic_serbian())
    'TEST<MILICA'
    >>> transliterate("Þĩŝ ïŜ Á ţęšť", dictionary_.latin_based(), " ")
    'THIS IS A TEST'
    >>> transliterate("محمود عبدالرحيم", dictionary_.arabic(), "-")
    'MXHMWD-EBDALRXHYM'
    >>> transliterate("παράδειγμα δοκιμής", dictionary_.greek())
    'PAPADEIGMA<DOKIMIS'
    """
    word = string.replace(u"\u002D", u"\u0020").split(u"\u0020")
    for i in range(len(word)):
        final_word = ""
        for char in word[i]:
            final_word += dictionary[char] if char in dictionary else char
        word[i] = final_word.upper()
    return sep.join(word)


def get_doc(cls):
    name = cls.__class__.__name__
    return name[:name.find("Code")]


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)




# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

# TODO: Shorten
titles = [
    "MR",
    "SIR",
    "JR",
    "MRS",
    "MISS",
    "MS",
    "MADAM",
    "DAME",
    "DR",
    "LADY",
    "LORD",
    "MX",
    "ESQ",
    "HE",
    "HON",
    "PROF",
    "MEP",
    "MP",
    "BT",
    "HON",
    "QC",
    "KC",
    "HH",
    "HAH",
    "HE",
    "HMEH",
    "REVD",
    "REV",
    "FR",
    "PR",
    "BR",
    "IMAN",
    "MUFTI",
    "HAFIZ",
    "HAFIZAH",
    "QARI",
    "MAWLANA",
    "HAJI",
    "SAYYID",
    "SAYYIDAH",
    "SHARIF",
    "SRI",
    "SREE",
    "SHRI",
    "SHREE",
    "SIRI",
    "SERI",
    "SMT",
    "KUM",
    "HER",
    "FRAU",
    "SIE",
    "EXCMO",
    "ILMO"
]



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned
# on making available complete source code of licensed works
# and modifications, which include larger works using a licensed
# work, under the same license. Copyright and license notices
# must be preserved. Contributors provide an express grant of
# patent rights.
#
# For more information on this, and how to apply and follow the
# GNU GPL, see http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018


def arabic():
    from mrz.generator.dictionaries.arabic import transliteration
    return transliteration


def cyrillic():
    from mrz.generator.dictionaries.cyrillic import transliteration
    return transliteration


def cyrillic_belarussian():
    from mrz.generator.dictionaries.cyrillic_belarussian import transliteration
    return transliteration


def cyrillic_bulgarian():
    from mrz.generator.dictionaries.cyrillic_bulgarian import transliteration
    return transliteration


def cyrillic_macedonian():
    from mrz.generator.dictionaries.cyrillic_macedonian import transliteration
    return transliteration


def cyrillic_serbian():
    from mrz.generator.dictionaries.cyrillic_serbian import transliteration
    return transliteration


def cyrillic_ukrainian():
    from mrz.generator.dictionaries.cyrillic_ukrainian import transliteration
    return transliteration


def greek():
    from mrz.generator.dictionaries.greek import transliteration
    return transliteration


def latin_based():
    from mrz.generator.dictionaries.latin_based import transliteration
    return transliteration





# -*- coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# Iván Rincón 2018

from string import ascii_letters
from .errors import *
from .functions import full_capitalize, get_doc

import mrz.base.countries as countries


def date(string):
    """
    >>> date("001122")
    '001122'
    >>> date("200229")
    '200229'
    """
    if _is_string(string):
        try:
            from datetime import datetime
            datetime.strptime(string, "%y%m%d").strftime("%y%m%d")
        except ValueError:
            raise DateError(cause=string)
        else:
            return string


def sex(string):
    """
    >>> sex("m")
    'M'
    """
    if _is_string(string) and len(string) != 1 or string not in "MmFfXx":
        raise SexError(cause=string)
    return string.upper()


def field(string: str, str_length: int, field_description: str, exception="") -> str:
    """
    >>> field("string", 8, "description")
    'STRING<<'
    """
    if _is_string(string) and len(string) > str_length:
        raise LengthError(document=field_description, cause=len(string), length=str_length, amx=True)
    for c in string:
        if c not in ascii_letters + "0123456789" + exception:
            raise FieldError(cause=c)
    return string.upper().ljust(str_length, "<")


def country(string, dictionary=countries.english):
    """
    >>> country("alb")
    'ALB'
    >>> country("yemen")
    'YEM'
    """
    if _is_string(string) and string.upper() in dictionary.values():
        return string.upper().ljust(3, "<")
    elif full_capitalize(string) in dictionary.keys():
        return dictionary[full_capitalize(string)].ljust(3, "<")
    else:
        raise CountryError(cause=string)


def document_type(string, cls):
    ok = False
    doc = get_doc(cls)
    s = string.upper()
    if _is_string(s) and s and len(s) <= 2:
        if doc == "TD1" or doc == "TD2":
            if s[0] in "IiAaCc" and s.find("V") != 1 and s != "AC":
                ok = True
        elif doc == "TD3" or doc == "Passport":
            if s[0] in "Pp":
                ok = True
        elif doc == "MRVA" or doc == "MRVB":
            if s[0] in "Vv":
                ok = True
    if not ok:
        raise DocumentTypeError(cause=string)
    return s.ljust(2, "<")


def precheck(document_description: str, string: str, length: int):
    s = string.replace("\n", "")
    if _is_string(_check_upper(string)) and len(s) != length:
        raise LengthError(cause=len(s), document=document_description, length=length)
    if not is_printable(s):
        raise FieldError("%s contains invalid characters" % document_description, s)


def is_printable(string: str) -> bool:
    for c in string:
        if c not in "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789<":
            return False
    return True


def uses_nums(string: str) -> bool:
    for c in string:
        if c in "0123456789":
            return True
    return False


def begin_by(string: str, char: str) -> bool:
    if string[0] != char:
        return False
    return True


def _check_upper(string):
    if string.isupper() or string.replace("<", "0").isdigit():
        return string
    else:
        raise FieldError("String contains invalid characters. Must be upper.", string)


def is_empty(string, padding="<"):
    return string == padding * len(string)


def _is_string(o):
    if isinstance(o, str):
        return True
    else:
        raise TypeError("Invalid argument. It should be a string, not a %s" % str(type(o).__name__))


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)




# -*- coding: utf-8 -*-
#
# GNU General Public License v3.0
#
# Iván Rincón 2018

from ..base.functions import transliterate


class _HolderName:
    def __init__(self, surname: str, given_names: str, transliteration: dict):
        self.transliteration = transliteration
        self.surname = surname
        self.given_names = given_names

    @property
    def transliteration(self) -> dict:
        """Return the transliteration dictionary used for non-ascii characters

        """
        return self._transliteration

    @transliteration.setter
    def transliteration(self, value: dict):
        """Set the transliteration dictionary used for non-ascii characters

        National characters may be used in the visual inspection zone for identifiers. If the national
        characters are not ASCII, a transliteration into Latin ASCII chars is provided by the
        transliteration methods included in transliteration.py or another transliteration dict.


        """
        self._transliteration = value

    @property
    def surname(self) -> str:
        """Return the primary identifier, usually the holder's surname.

        The issuing State or organization shall establish which part of the name is the primary identifier.
        This may be the family name, the maiden name or the married name, the main name, the surname, and
        in some cases, the entire name where the holder’s name cannot be divided into two parts.


        """
        return self._surnames

    @surname.setter
    def surname(self, value: str):
        """Set the primary identifier, usually the holder's surname

        The issuing State or organization shall establish which part of the name is the primary identifier.
        This may be the family name, the maiden name or the married name, the main name, the surname, and
        in some cases, the entire name where the holder’s name cannot be divided into two parts.

        Case insensitive.

        """
        self._surnames = transliterate(value, self.transliteration)

    @property
    def given_names(self) -> str:
        """Return the secondary identifier, usually holder's given names.

        The remaining parts of the name are the secondary identifier. These may be the forenames, familiar
        names, given names, initials, or any other secondary names.

        """
        return self._given_names

    @given_names.setter
    def given_names(self, value: str):
        """Set the secondary identifier, usually holder's given names.

        The remaining parts of the name are the secondary identifier. These may be the forenames, familiar
        names, given names, initials, or any other secondary names.

        Case insensitive.

        """
        self._given_names = transliterate(value, self.transliteration)

    def __str__(self):
        return self.surname + "<<" + self.given_names


# -*- Coding: UTF-8 -*-
#
# GNU General Public License v3.0
#
# Permissions of this strong copyleft license are conditioned on making available
# complete source code of licensed works and modifications, which include larger works
# using a licensed work, under the same license. Copyright and license notices must be
# preserved. Contributors provide an express grant of patent rights.
#
# For more information on this, and how to apply and follow theGNU GPL, see:
# http://www.gnu.org/licenses
#
# (ɔ) Iván Rincón 2018

english = {
    "Afghanistan": "AFG",
    "Aland Islands": "ALA",
    "Albania": "ALB",
    "Algeria": "DZA",
    "American Samoa": "ASM",
    "Andorra": "AND",
    "Angola": "AGO",
    "Anguilla": "AIA",
    "Antarctica": "ATA",
    "Antigua And Barbuda": "ATG",
    "Argentina": "ARG",
    "Armenia": "ARM",
    "Aruba": "ABW",
    "Australia": "AUS",
    "Austria": "AUT",
    "Azerbaijan": "AZE",
    "Bahamas": "BHS",
    "Bahrain": "BHR",
    "Bangladesh": "BGD",
    "Barbados": "BRB",
    "Belarus": "BLR",
    "Belgium": "BEL",
    "Belize": "BLZ",
    "Benin": "BEN",
    "Bermuda": "BMU",
    "Bhutan": "BTN",
    "Bolivia":  "BOL",
    "Bonaire, Sint Eustatius And Saba": "BES",
    "Bosnia And Herzegovina": "BIH",
    "Botswana": "BWA",
    "Bouvet Island": "BVT",
    "Brazil": "BRA",
    "British Indian Ocean Territory": "IOT",
    "Brunei": "BRN",
    "Bulgaria": "BGR",
    "Burkina Faso": "BFA",
    "Burundi": "BDI",
    "Cabo Verde":  "CPV",
    "Cambodia": "KHM",
    "Cameroon": "CMR",
    "Canada": "CAN",
    "Cayman Islands": "CYM",
    "Central African Republic": "CAF",
    "Chad": "TCD",
    "Chile": "CHL",
    "China": "CHN",
    "Christmas Island": "CXR",
    "Cocos Islands": "CCK",
    # "Keeling Islands": "CCK",  # *
    "Colombia": "COL",
    "Comoros": "COM",
    "Congo": "COG",
    "Democratic Republic Of The Congo": "COD",
    "Cook Islands": "COK",
    "Costa Rica": "CRI",
    "Ivory Coast": "CIV",
    "Croatia": "HRV",
    "Cuba": "CUB",
    "Curaçao": "CUW",
    "Cyprus": "CYP",
    "Czech Republic": "CZE",
    "Denmark": "DNK",
    "Djibouti": "DJI",
    "Dominica": "DMA",
    "Dominican Republic ": "DOM",
    "Ecuador": "ECU",
    "Egypt": "EGY",
    "El Salvador": "SLV",
    "Equatorial Guinea": "GNQ",
    "Eritrea": "ERI",
    "Estonia": "EST",
    "Ethiopia": "ETH",
    "Falkland Islands": "FLK",  # Sovereignty dispute between Argentina and United Kingdom
    "Faroe Islands": "FRO",
    "Fiji": "FJI",
    "Finland":  "FIN",
    "France": "FRA",
    "French Guiana": "GUF",
    "French Polynesia": "PYF",
    "French Southern Territories":  "ATF",
    "Gabon": "GAB",
    "Gambia": "GMB",
    "Georgia": "GEO",
    "Germany": "D",
    "Ghana": "GHA",
    "Gibraltar": "GIB",
    "Greece": "GRC",
    "Greenland": "GRL",
    "Grenada": "GRD",
    "Guadeloupe": "GLP",
    "Guam": "GUM",
    "Guatemala": "GTM",
    "Guernsey": "GGY",
    "Guinea": "GIN",
    "Guinea Bissau": "GNB",
    "Guyana": "GUY",
    "Haiti": "HTI",
    "Heard Island And Mcdonald Islands": "HMD",
    # "Holy See": "VAT",
    "Vatican City State": "VAT",  # *
    "Honduras": "HND",
    "Hong Kong": "HKG",
    "Hungary": "HUN",
    "Iceland": "ISL",
    "India": "IND",
    "Indonesia": "IDN",
    "Iran": "IRN",
    "Iraq": "IRQ",
    "Ireland": "IRL",
    "Isle Of Man": "IMN",
    "Israel": "ISR",
    "Italy": "ITA",
    "Jamaica": "JAM",
    "Japan": "JPN",
    "Jersey": "JEY",
    "Jordan": "JOR",
    "Kazakhstan": "KAZ",
    "Kenya": "KEN",
    "Kiribati": "KIR",
    "North Korea": "PRK",
    "South Korea": "KOR",
    "Kuwait": "KWT",
    "Kyrgyzstan": "KGZ",
    "Lao Democratic Republic": "LAO",
    "Latvia": "LVA",
    "Lebanon": "LBN",
    "Lesotho": "LSO",
    "Liberia": "LBR",
    "Libya": "LBY",
    "Liechtenstein": "LIE",
    "Lithuania": "LTU",
    "Luxembourg": "LUX",
    "Macao": "MAC",
    "Macedonia": "MKD",
    "Madagascar": "MDG",
    "Malawi": "MWI",
    "Malaysia": "MYS",
    "Maldives": "MDV",
    "Mali": "MLI",
    "Malta": "MLT",
    "Marshall Islands": "MHL",
    "Martinique": "MTQ",
    "Mauritania": "MRT",
    "Mauritius": "MUS",
    "Mayotte": "MYT",
    "Mexico": "MEX",
    "Micronesia": "FSM",
    "Moldova": "MDA",
    "Monaco": "MCO",
    "Mongolia": "MNG",
    "Montenegro": "MNE",
    "Montserrat": "MSR",
    "Morocco": "MAR",
    "Mozambique": "MOZ",
    "Myanmar": "MMR",
    "Namibia": "NAM",
    "Nauru": "NRU",
    "Nepal": "NPL",
    "Netherlands": "NLD",
    "Netherlands Antilles": "ANT",
    "Neutral Zone": "NTZ",
    "New Caledonia": "NCL",
    "New Zealand": "NZL",
    "Nicaragua": "NIC",
    "Niger": "NER",
    "Nigeria": "NGA",
    "Niue": "NIU",
    "Norfolk Island": "NFK",
    "Northern Mariana Islands": "MNP",
    "Norway": "NOR",
    "Oman": "OMN",
    "Pakistan": "PAK",
    "Palau": "PLW",
    "Palestine": "PSE",
    "Panama": "PAN",
    "Papua New Guinea": "PNG",
    "Paraguay": "PRY",
    "Peru": "PER",
    "Philippines": "PHL",
    "Pitcairn": "PCN",
    "Poland": "POL",
    "Portugal": "PRT",
    "Puerto Rico": "PRI",
    "Qatar": "QAT",
    "Reunion": "REU",
    "Romania": "ROU",
    "Russia": "RUS",
    "Rwanda": "RWA",
    "Saint Barthelemy": "BLM",
    "Saint Helena, Ascension And Tristan Da Cunha": "SHN",
    "Saint Kitts And Nevis": "KNA",
    "Saint Lucia": "LCA",
    "Saint Martin": "MAF",
    "Saint Pierre And Miquelon": "SPM",
    "Saint Vincent And The Grenadines": "VCT",
    "Samoa": "WSM",
    "San Marino": "SMR",
    "Sao Tome And Principe": "STP",
    "Saudi Arabia": "SAU",
    "Senegal": "SEN",
    "Serbia": "SRB",
    "Seychelles": "SYC",
    "Sierra Leone": "SLE",
    "Singapore": "SGP",
    "Sint Maarten": "SXM",
    "Slovakia": "SVK",
    "Slovenia": "SVN",
    "Solomon Islands": "SLB",
    "Somalia": "SOM",
    "South Africa": "ZAF",
    "South Georgia And The South Sandwich Islands": "SGS",
    "South Sudan": "SSD",
    "Spain": "ESP",
    "Sri Lanka": "LKA",
    "Sudan": "SDN",
    "Suriname": "SUR",
    "Svalbard And Jan Mayen": "SJM",
    "Swaziland": "SWZ",
    "Sweden": "SWE",
    "Switzerland": "CHE",
    "Syria": "SYR",
    "Taiwan": "TWN",
    "Tajikistan": "TJK",
    "Tanzania": "TZA",
    "Thailand": "THA",
    "East Timor": "TLS",
    "Togo": "TGO",
    "Tokelau": "TKL",
    "Tonga": "TON",
    "Trinidad And Tobago": "TTO",
    "Tunisia": "TUN",
    "Turkey": "TUR",
    "Turkmenistan": "TKM",
    "Turks And Caicos Islands": "TCA",
    "Tuvalu": "TUV",
    "Uganda": "UGA",
    "Ukraine": "UKR",
    "Arab Emirates": "ARE",
    "United Kingdom": "GB",
    "British": "GBR",
    "British Overseas Territories": "GBD",
    "British National Overseas": "GBN",
    "British Overseas Citizen": "GBO",
    "British Protected": "GBP",
    "British Subject": "GBS",
    "United States": "USA",
    "United States Minor Outlying Islands": "UMI",
    "Uruguay": "URY",
    "Utopia": "UTO",  # ICAO Example. Not real country
    "Uzbekistan": "UZB",
    "Vanuatu": "VUT",
    "Venezuela": "VEN",
    "Vietnam": "VNM",
    "Virgin Islands": "VGB",
    "Virgin Islands Of The United States": "VIR",
    "Wallis And Futuna": "WLF",
    "Western Sahara": "ESH",
    "Yemen": "YEM",
    "Zambia": "ZMB",
    "Zimbabwe": "ZWE",
    "European Union": "EUE",
    "United Nations Organization": "UNO",
    "United Nations Agency": "UNA",
    "Kosovo": "UNK",
    "African Development Bank": "XBA",
    "Afrexim": "XIM",
    "Caricom": "XCC",
    "Comesa": "XCO",
    "Ecowas": "XEC",
    "Interpol": "INP",
    "Military Order Of Malta": "XOM",
    "Stateless": "XXA",
    "Refugee": "XXA",
    "Refugee B": "XXB",
    "Unknown": "XXX"
}
